﻿//
// Copyright (c) 2003-2007, YaPB Development Team. All rights reserved.
//
// Redistribution and use in source and binary forms, with or without 
// modification, are permitted provided that the following conditions are 
// met:
//
// Redistributions of source code must retain the above copyright notice, 
// this list of conditions and the following disclaimer.
//
// Redistributions in binary form must reproduce the above copyright 
// notice, this list of conditions and the following disclaimer in the 
// documentation and/or other materials provided with the distribution.
//
// Neither the name of the YaPB Development Team nor the names of its 
// contributors may be used to endorse or promote products derived from 
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS 
// IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED 
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
// PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT 
// OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
// SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 
// LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
// THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// Author(s): Dmitry Zhukov (jeefo@inbox.ru).
//

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.Collections;
using System.Windows.Forms.VisualStyles;
using System.Runtime.InteropServices;
using System.Globalization;

namespace ConfCenter
{
   public partial class Main: Form
   {

      public const int NUM_WEAPONS = 26;
      public const int NUM_PERSONALITIES = 4;
      public const int NUM_SKILLENTRIES = 8;

      public static readonly string VERSION = "1.4";
      public static readonly string CAPTION = "YaPB ConfCenter Application";
      public static readonly string CONFIG_HDR = "// ======================================================================== //\n//  This is weapon configuration file for Yet Another POD-Bot.\n//  Copyright (c) 2003-2007, by YaPB Development Team.\n// ------------------------------------------------------------------------ //\n//  Any line begining with '//' in the first column is a Comment Line.\n//  Please check documentation file for more information about config.\n// ------------------------------------------------------------------------ //\n//  Строки, которые начинаются с '//' - это первые символы комментария.\n//  Пожалуйста просмотрите файл справки для большей инфы о этом конфиге.\n// ======================================================================== //\n";

      static public void WriteFile_ANSI (string file, string contents)
      {
         if ((File.GetAttributes (file) & FileAttributes.ReadOnly) == FileAttributes.ReadOnly)
            File.SetAttributes (file, File.GetAttributes (file) ^ FileAttributes.ReadOnly);

         using (TextWriter write = new StreamWriter (file, false, Encoding.Default))
         {
            write.Write (contents);
         }
      }

      public class Localizer
      {
         public class LangEntry
         {
            public string orig;
            public string tran;

            public LangEntry (string o, string t)
            {
               orig = o;
               tran = t;
            }
         };

         static public Hashtable tab;
         static public bool localOK;

         public Localizer ()
         {
            tab = new Hashtable ();

            tab.Add ("Determines if one slot is keep free for player connection.", "Определяет, сохраняется ли свободным один слот на сервере для присоединения человека.");
            tab.Add ("Determines number of bots playing on the server machine.", "Определяет количество ботов которые будет игать на сервере.");
            tab.Add ("Determines minimum bots skill, when creating bot with random parameters.", "Определяет минимальный скилл ботов, когда бота присоединяют со случайными параметрами.");
            tab.Add ("Determines maximum bots skill, when creating bot with random parameters.", "Определяет максимальный скилл ботов, когда бота присоединяют со случайными параметрами.");
            tab.Add ("Determines number of bots that can follow player, that used Follow me radio message.", "Определяет максимальное количество ботов, которые могут следовать за товарищем который использовал радио Follow Me.");
            tab.Add ("Determines time interval, used with bot hearing system.", "Определяет интервал времени используемый для проверки слуха ботов.");
            tab.Add ("Determines time interval, used to check for items around the bot.", "Определяет интервал времени используемый для поиска различных предметов вокруг бота.");
            tab.Add ("Determines time interval, used to check for grenade throw.", "Определяет интервал времени используемый для проверки кидания гранат.");
            tab.Add ("Determines if bots able to chat each-other while they are dead.", "Определяет, могут ли боты общаться друг с другом, через сообщения чата, когда они мертвы.");
            tab.Add ("Determines if bot dll able to comment some of the commands, by halflife speech engine.", "Определяет, может ли библиотека бота комментировать некоторые команды движком голоса халфлайф.");
            tab.Add ("Determines mode, where bots able to fight only on knives.", "Определяет, будут ли боты сражаться только на ножах или нет.");
            tab.Add ("Determines, if skill value for bot will be appended to it name, after bot creation.", "Определяет, будет ли дописан скилл бота, сразу же после его имени во время создания бота.");
            tab.Add ("Determines, if all bots on the server stop function, while this variable set to 1.", "Определяет, будут ли все боты на сервере полностью остановлены.");
            tab.Add ("Determines, if bots able to shoot through wall when suspectiong an enemy.", "Определяет, могут ли боты стрелять сквозь стены когда кого либо слышит, либо когда предполагает, что там есть враг.");
            tab.Add ("Determines, if bots can vote for kick, against it's teamkiller, or for map change.", "Определяет, могут ли боты голосовать против своих тимкиллеров, и за различные карты.");
            tab.Add ("Determines, if bots are able to spray graphical logos in a game.", "Определяет, могут ли боты рисовать логотипы на земли на старте раунда и у главных точек карты.");
            tab.Add ("Determines, if bots able to purchase weapon and items at start of the each round.", "Определяет, могут ли боты закупаться оружием и предметами в начале каждого раунда.");
            tab.Add ("Determines level for bot debugging engine. For developers and betatesting team only.", "Определяет режим отладки для ботов (используется только для разработки и бета-тестировании).");
            tab.Add ("Determines method for aiming of bots.", "Определяет режим поворотов и нацеливания ботов.");
            tab.Add ("Determines prefix which will be added to start of the bot name, at the creation time.", "Определяет приставку (клан) который будет добавлен перед именем бота во время его создания.");
            tab.Add ("Determines the language in which all bot menus and comments are displayed.", "Определяет язык используемый для отображения меню и сообщений ботов.");
            tab.Add ("Determines password for accessing bot menu and commands on the server running yapb.", "Определяет пароль используемый для доступа к меню и командам ботов с клиента сервера, который использует yapb.");
            tab.Add ("Determines pasword key for setinfo information to acccess to the bot menu and command.", "Определяет ключ к паролю для setinfo на сервере.");
            tab.Add ("Determines path to the Official CSBot's chatter files.", "Определяет путь к файлам чаттера бота (либо Official CSBot Chatter, либо Reallite CSBot Chatter).");
            tab.Add ("Determines if bot's teamkiller will be punshied for his heroic action.", "Определяет, могут ли боты мстить напарникам за нанесение ущерба здаровью, и убивает тимкиллера.");
            tab.Add ("Determines method for communication engine for bots.", "Определяет движок используемый для общения ботов во время игры.");
            tab.Add ("Determines if bots are able to use team-economics to save money.", "Определяет, могут ли боты использовать командную эакономику.");
            tab.Add ("Determines the team, in which all creating bots will be forced to join.", "Определяет команду к которой боты должны присоединяться в начале игры.");
            tab.Add ("Determines list of the weapons, that are banned for buying and picking up.", "Определяет список оружиея, который бот не может покупать и поднимать с земли.");
            tab.Add ("Determines danger factor for bots experience system, higher values resulting more tactical bots.", "Определяет фактор опастности для алгоритма поиска путей. Более высокие значения делают ботов более умными.");
            tab.Add ("Determines if bots are able to shoot their enemies.", "Определяет, не могут ли боты стрелять по своей цели или могут.");
            tab.Add ("Determines if bots can only be running in hard mode. Enable only if you playing very well.", "Определяет, будут ли боты работать в режиме повышенной сложности или нет.");
            tab.Add ("Determines percent of bots that can autofollow human player at start of each round.", "Определяет процент ботов, которые могут автоматически следовать за человеком в начале раунда.");
            tab.Add ("YaPB ConfCenter Application", "Конфиг Центр YaPB");
            tab.Add ("Minimum reaction delay", "Мин. задержка реакции");
            tab.Add ("Maximum reaction delay", "Макс. задержка реакции");
            tab.Add ("Minimum turning speed", "Мин. скорость поворота");
            tab.Add ("Maximum turning speed", "Макс. скорость поворота");
            tab.Add ("Allowed headshot percent", "Процент хэдшотов");
            tab.Add ("Percent shooting on hearing", "Процент стрельбы на слух");
            tab.Add ("Percent shooting on seeing", "Процет стрельбы на глаз");
            tab.Add ("Recoil compensation value", "Значение отдачи оружия");
            tab.Add ("Definition for: ", "Настройка для: ");
            tab.Add ("YaPB Installation not found in registry. Do you want to show directory youself ?", "Папка установки YaPB не найдена. Хотите найти ее сами ?");
            tab.Add ("Select root YaPB Path.", "Выберите путь ка папке YaPB.");
            tab.Add ("You're selected wrong directory. Application Terminates.", "Вы выбрали не корректную папку. Прога Отрубается.");
            tab.Add ("Cannot move item to this position. Index is out of range.", "Невозможно передвинуть пушку на эту позицию. Индекс вышел за допустимые пределы.");
            tab.Add ("Are you sure you want reset skill settings ?", "Ты уверен что хочешь скинуть настройки скилла на стандарт ?");
            tab.Add ("[ru] Russian Federation  (Russia)", "[ru] Россия (Русский)");
            tab.Add ("[en] United States (English)", "[en] США (Английский)");
            tab.Add ("[chs] China (Simplified Chinese)", "[chs] Китай (Упрощенный Китайский)");
            tab.Add ("[pl] Poland (Polish)", "[pl] Польша (Польский)");
            tab.Add ("\r\nValue for ", "\r\nЗначение для ");
            tab.Add ("Enabled", "Включено");
            tab.Add ("Disabled", "Отключено");
            tab.Add ("None", "Нет");
            tab.Add (" entry is not valid. Please fix config.", " - критическая ошибка в секции. Невозможно обработать конфиг.");
            tab.Add ("Unaccpetable values in bot economics configuration. Defaults are set.", "Недопустимые значения в конфигурации экономики. Умолчания установлены.");
            tab.Add ("Writing configs...", "Записываю Конфиги...");
            tab.Add ("Something weird, with your config. Probably some values are messed. Config loaded not completely.\n\nException:\n", "Что-то странное твориться с вашим конфигом (основным). Возможно испорчены значения. Конфиг загружен не польностью.\n\nДамп:\n");
            tab.Add ("Knife priority cannot be changed.", "Приоритет ножа не может быть изменен.");
            tab.Add ("Cannot move item to knife position. Knife priority cannot be changed.", "Невозможно установить оружие на позицию ножа. Приоритет ножа не может быть изменен.");
            tab.Add ("Determines the list of logos used for bot spray logo action.", "Определяет список логотипов использумых ботом.");
            tab.Add ("Determines method for calculating the msec value.", "Определяет метод просчета врмени кадра. В зависимости от компьютера, влияет на частость застревания бота.");
				tab.Add ("Determines log level for bots messages. Critical errors will be logged anyway.", "Определяет что бот будет писать в логи. Критические ошибки записываются всегда.");
			}

         public static string T (string input)
         {
            if (!Localizer.localOK || !tab.Contains (input))
               return input;

            return tab[input].ToString ();
         }
      };

      public enum Personality: int
      {
         Undefined = 0,
         Rusher,
         Careful,
         Normal
      };

      public enum SkillLevel: int
      {
         Skill0 = 0,
         Skill20,
         Skill40,
         Skill60,
         Skill80,
         Skill100
      };


      public sealed class WeaponItem
      {
         private int id;
         private int order;
         private string desc;
         private string name;

         public WeaponItem (int id, string name, string desc)
         {
            this.id = id;
            this.order = -1;
            this.desc = desc;
            this.name = name;
         }

         public int Id
         {
            get { return id; }
            set { id = value; }
         }
         
         public string Name
         {
            get { return name; }
            set { name = value; }
         }
         
         public string Desc
         {
            get { return desc; }
            set { desc = value; }
         }
         
         public int Order
         {
            get { return order; }
            set { order = value; }
         }
      };

      public sealed class SkillItem
      {
         public double minDelay;
         public double maxDelay;
         public double minTurnspeed;
         public double maxTurnspeed;
         public int headshotAllow;
         public int hearShootThru;
         public int seenShootThru;
         public int recoilAmount;

         public SkillItem (double minDelay, double maxDelay, double minTurnspeed, double maxTurnspeed, int headshotAllow, int hear, int seen, int recoil)
         {
            this.minDelay = minDelay;
            this.maxDelay = maxDelay;
            this.minTurnspeed = minTurnspeed;
            this.maxTurnspeed = maxTurnspeed;
            this.headshotAllow = headshotAllow;
            this.hearShootThru = hear;
            this.seenShootThru = seen;
            this.recoilAmount = recoil;
         }

         public SkillItem ()
         {
            this.minDelay = 0;
            this.maxDelay = 0;
            this.minTurnspeed = 20;
            this.maxTurnspeed = 20;
            this.headshotAllow = 0;
            this.hearShootThru = 0;
            this.seenShootThru = 0;
            this.recoilAmount = 20;
         }
      };

      public sealed class GeneralManager
      {
         private TreeView treeControl;

         static public readonly SkillItem[] defaultData =
         {
           new SkillItem (0.8, 1.0, 2.0, 3.0,  0, 0, 0, 50), 
           new SkillItem (0.6, 0.8, 3.0, 4.0,  10, 0, 0, 40),
           new SkillItem (0.4, 0.6, 4.0, 6.0,  30, 0, 50, 35), 
           new SkillItem (0.2, 0.3, 6.0, 8.0,  0, 30, 80, 30),
           new SkillItem (0.1, 0.2, 8.0, 10.0, 80, 50, 100, 23), 
           new SkillItem (0.0, 0.1, 9.0, 12.0, 100, 100, 100, 20)
         };

         public SkillItem[] loadData = defaultData;

         static public readonly string[] descriptionList =
         {
            Localizer.T ("Minimum reaction delay"),
            Localizer.T ("Maximum reaction delay"),
            Localizer.T ("Minimum turning speed"),
            Localizer.T ("Maximum turning speed"),
            Localizer.T ("Allowed headshot percent"),
            Localizer.T ("Percent shooting on hearing"),
            Localizer.T ("Percent shooting on seeing"),
            Localizer.T ("Recoil compensation value")
         };

         static public readonly string[] configValues =
         {
            "MIN_DELAY",
            "MAX_DELAY",
            "MIN_TURNSPEED",
            "MAX_TURNSPEED",
            "HEADSHOT_ALLOW",
            "HEAR_SHOOTTHRU",
            "SEEN_SHOOTTHRU",
            "RECOIL_AMOUNT"
         };

         private ListView listControl;
         private ComboBox comboCtl;
         private ArrayList[] weaponList;

         public static int[] Economics;
         public static int[] Percentage;

         internal void LoadTimeSetOrderForId (Personality per, string id, int order)
         {
            for (int i = 0; i < weaponList[(int)per].Count; i++)
            {
               WeaponItem item = GetWeaponById (per, i);

               if (item.Id == Convert.ToInt32 (id))
                  item.Order = order;
            }
         }

         internal string GetPersonalityTab (Personality per)
         {
            int id = 0;

            if (per == Personality.Rusher)
               id = 0;
            else if (per == Personality.Normal)
               id = 1;
            else if (per == Personality.Careful)
               id = 2;

            comboCtl.SelectedIndex = id;
            RebuildTable (per);

            StringBuilder builer = new StringBuilder ();

            OrderSort os = new OrderSort ();
            weaponList[(int)per].Sort (os);

            for (int i = 0; i < weaponList[(int)per].Count; i++)
            {
               WeaponItem item = GetWeaponById (per, i);

               if (Convert.ToString (item.Id).Length == 1)
                  builer.Append ('0');

               builer.Append (item.Id);

               if (i != weaponList[(int)per].Count - 1)
                  builer.Append (",");

            }
            return builer.ToString ();
         }

         public void WriteOutputConfig (string toFile)
         {
            // info:
            // since first two entries are fake, we write manually, since new restriction system, but we leave it there. :)
            string configHead = CONFIG_HDR + "\nMapStandard = -1,0,-1,2,-1,0,1,2,2,2,-1,2,-1,-1,0,0,1,0,1,1,2,2,0,1,2,1\nMapAS = -1,-1,-1,2,-1,0,1,1,1,1,1,1,0,2,0,-1,1,0,1,1,0,0,-1,1,1,1\n\n";
 

            configHead += "GrenadePercent = " + GeneralManager.Percentage[0] + "," + GeneralManager.Percentage[1] + "," + GeneralManager.Percentage[2];
            configHead += "\nEconomics = ";

            for (int i = 0; i < 11; i++)
            {
               configHead += GeneralManager.Economics[i];

               if (i != 10)
                  configHead += ",";
            }

            string rusher = GetPersonalityTab (Personality.Rusher);
            string careful = GetPersonalityTab (Personality.Careful);
            string normal = GetPersonalityTab (Personality.Normal);

            configHead += "\n\nPersonalityNormal = " + normal + "\nPersonalityRusher = " + rusher + "\nPersonalityCareful = " + careful + "\n";

           	int dummy = 0;

            for (int i = 0; i < 6; i++)
            {
               SkillLevel level = (SkillLevel)i;

               string dataType = "";
               GenerateNodeData (level, ref dataType, ref dummy);

               configHead += "\nSkill" + dataType + " = ";

               for (int j = 0; j < NUM_SKILLENTRIES; j++)
               {
                  if (configValues[j] == "MIN_TURNSPEED" || configValues[j] == "MAX_TURNSPEED")
                     configHead += GetValue (level, configValues[j]) + ".0,";
                  else if (configValues[j] == "MAX_DELAY" || configValues[j] == "MIN_DELAY")
                  {
                     string val = (Convert.ToSingle (GetValue (level, configValues[j])) / 100).ToString ().Replace (',', '.');

                     if (val.Length == 1 || val.Length == 3)
                        val += "0";

                     if (val == "00")
                        val = "0.0";
                     else if (val == "10")
                        val = "1.0";

                     configHead += val + ",";
                  }
                  else
                     configHead += GetValue (level, configValues[j]) + ",";
               }
               configHead = configHead.Remove (configHead.Length - 1, 1);
             //  configHead += "\n";
            }

            WriteFile_ANSI (toFile, configHead);
         }


         public void LoadData ()
         {
            if (loadData == null)
            {
               loadData = new SkillItem[7];

               for (int i = 0; i < 7; i++)
                  loadData[i] = new SkillItem ();
            }

            string cfg = configPath + "general.cfg";

            using (TextReader reader = new StreamReader (File.OpenRead (cfg)))
            {
               for (; ; )
               {
                  string line = reader.ReadLine ();

                  if (line == null)
                     break;

                  if ((line.Contains ("//") && line[0] == '/' && (line[1] == '/')) || line == "\n" || line == "\r" || line == "\0")
                     continue;

                  line.Trim ('\n', '\r');

                  string[] data = line.Split ('=');

                  if (data.Length < 2)
                     continue;

                  data[0] = data[0].Trim ().Trim ();
                  data[1] = data[1].Trim ().Trim ();

                  string[] split = data[1].Split (',');

                  if (data[0] == "GrenadePercent")
                  {
                     if (split.Length != 3)
                        throw new Exception (data[0]);

                     for (int i = 0; i < 3; i++)
                        GeneralManager.Percentage[i] = Convert.ToInt32 (split[i]);
                  }
                  else if (data[0] == "Economics")
                  {
                     if (split.Length != 11)
                        throw new Exception (data[0]);

                     for (int i = 0; i < 11; i++)
                        GeneralManager.Economics[i] = Convert.ToInt32 (split[i]);
                  }
                  else if (data[0] == "PersonalityNormal")
                  {
                     if (split.Length != NUM_WEAPONS)
                        throw new Exception (data[0]);

                     for (int i = 0; i < NUM_WEAPONS; i++)
                        LoadTimeSetOrderForId (Personality.Normal, split[i], i);
                  }
                  else if (data[0] == "PersonalityRusher")
                  {
                     if (split.Length != NUM_WEAPONS)
                        throw new Exception (data[0]);

                     for (int i = 0; i < NUM_WEAPONS; i++)
                        LoadTimeSetOrderForId (Personality.Rusher, split[i], i);
                  }
                  else if (data[0] == "PersonalityCareful")
                  {
                     if (split.Length != NUM_WEAPONS)
                        throw new Exception (data[0]);

                     for (int i = 0; i < NUM_WEAPONS; i++)
                        LoadTimeSetOrderForId (Personality.Careful, split[i], i);
                  }
                  else if (data[0].Contains ("Skill"))
                  {
                     if (split.Length != 8)
                        throw new Exception (data[0]);

                     int skillParse = 0;

                     if (data[0].Contains ("Stupid"))
                        skillParse = 0;
                     else if (data[0].Contains ("Newbie"))
                        skillParse = 1;
                     else if (data[0].Contains ("Average"))
                        skillParse = 2;
                     else if (data[0].Contains ("Advanced"))
                        skillParse = 3;
                     else if (data[0].Contains ("Professional"))
                        skillParse = 4;
                     else if (data[0].Contains ("Godlike"))
                        skillParse = 5;

                     for (int i = 0; i < 8; i++)
                     {
                        switch (i)
                        {
                        case 0:
                           loadData[skillParse].minDelay = UnmanagedUtils.atof (split[i]) * 100;
                           break;

                        case 1:
                           loadData[skillParse].maxDelay = UnmanagedUtils.atof (split[i]) * 100;
                           break;

                        case 2:
                           loadData[skillParse].minTurnspeed = UnmanagedUtils.atof (split[i]);
                           break;

                        case 3:
                           loadData[skillParse].maxTurnspeed = UnmanagedUtils.atof (split[i]);
                           break;

                        case 4:
                           loadData[skillParse].headshotAllow = Convert.ToInt32 (split[i]);
                           break;

                        case 5:
                           loadData[skillParse].hearShootThru = Convert.ToInt32 (split[i]);
                           break;

                        case 6:
                           loadData[skillParse].seenShootThru = Convert.ToInt32 (split[i]);
                           break;

                        case 7:
                           loadData[skillParse].recoilAmount = Convert.ToInt32 (split[i]);
                           break;
                        }
                     }
                  }
               }
            }
         }

         public void RegisterWeapon (string name, int id, string desc)
         {
            for (int i = 0; i < NUM_PERSONALITIES; i++)
               weaponList[i].Add (new WeaponItem (id, name, desc));
         }

         public WeaponItem GetWeaponById (Personality per, int id)
         {
            if ((id < 0) || (id >= weaponList[(int)per].Count))
               return null;

            return (WeaponItem)weaponList[(int)per][id];
         }

         public ListViewItem EntryToItem (Personality per, WeaponItem entry)
         {
            ListViewItem n = new ListViewItem ();
            ListViewItem.ListViewSubItem a = new ListViewItem.ListViewSubItem (n, "dumm");

            n.ImageIndex = entry.Id;
            n.ToolTipText = entry.Desc;

            n.SubItems.Insert (0, new ListViewItem.ListViewSubItem (n, entry.Name));
            n.SubItems.Insert (1, new ListViewItem.ListViewSubItem (n, Convert.ToString (entry.Id)));
            n.SubItems.Insert (2, new ListViewItem.ListViewSubItem (n, entry.Desc));

            return n;
         }

         public void RebuildTable (Personality forPer)
         {
            weaponList[(int)forPer].Clear ();

            for (int i = 0; i < listControl.Items.Count; i++)
               weaponList[(int)forPer].Add (ItemToEntry (forPer, listControl.Items[i], -1));
         }


         public WeaponItem ItemToEntry (Personality per, ListViewItem item, int newOrder)
         {
            ListViewItem.ListViewSubItemCollection coll = listControl.Items[item.Index].SubItems;

            WeaponItem n = new WeaponItem (Convert.ToInt32 (coll[1].Text), coll[0].Text, coll[2].Text);
            n.Order = item.Index;

            return n;
         }

         class OrderSort: IComparer
         {
            public int Compare (object a, object b)
            {
               if (((WeaponItem)a).Order < ((WeaponItem)b).Order)
                  return -1;

               return 0;
            }
         };

         public void Invalidate (Personality type)
         {
            OrderSort os = new OrderSort ();
            weaponList[(int)type].Sort (os);

            for (int i = 0; i < weaponList[(int)type].Count; i++)
               listControl.Items.Add (EntryToItem (type, (WeaponItem)weaponList[(int)type][i]));
         }

         public ArrayList Entries
         {
            get
            {
               return weaponList[(int)Personality.Normal];
            }
         }

         public GeneralManager (TreeView handle, ListView handler, ComboBox ctl)
         {
            treeControl = handle;
            listControl = handler;
            comboCtl = ctl;
            weaponList = new ArrayList[NUM_PERSONALITIES];

            for (int i = 0; i < 6; i++)
               AddSection ((SkillLevel)i);

            for (int i = 0; i < NUM_PERSONALITIES; i++)
               weaponList[i] = new ArrayList (NUM_WEAPONS);

            Economics = new int[NUM_WEAPONS];
            Percentage = new int[4];

            RegisterWeapon ("knife", 0, "Seal Knife");
            RegisterWeapon ("usp", 1, "Heckler & Koch Tactical USP .45");
            RegisterWeapon ("glock", 2, "Glock18 Select Fire");
            RegisterWeapon ("deagle", 3, "Desert Eagle .50AE");
            RegisterWeapon ("p228", 4, "Sig P228");
            RegisterWeapon ("elite", 5, "Dual Beretta 96G elites");
            RegisterWeapon ("fn57", 6, "Fabrique Nationale 5.7");
            RegisterWeapon ("m3", 7, "Benelli M3 Super90");
            RegisterWeapon ("xm1014", 8, "Benelli XM1014");
            RegisterWeapon ("mp5", 9, "Heckler & Koch MP5 Navy");
            RegisterWeapon ("tmp", 10, "Steyr Tactical Machine Pistol");
            RegisterWeapon ("p90", 11, "FabriqueNationale P90");
            RegisterWeapon ("mac10", 12, "Ingram MAC-10");
            RegisterWeapon ("ump45", 13, "Heckler & Koch UMP45");
            RegisterWeapon ("ak47", 14, "Avtomat Kalashnikov AK-47");
            RegisterWeapon ("sg552", 15, "Sig SG-552 Commando");
            RegisterWeapon ("m4a1", 16, "Colt M4A1 Carbine");
            RegisterWeapon ("galil", 17, "IMI Galil");
            RegisterWeapon ("famas", 18, "GIAT FAMAS");
            RegisterWeapon ("aug", 19, "Steyr Aug");
            RegisterWeapon ("scout", 20, "Steyr Scout");
            RegisterWeapon ("awp", 21, "Accuracy Int. Arctic Warfare/Magnum (AWM)");
            RegisterWeapon ("g3sg1", 22, "Heckler & Koch G3/SG-1");
            RegisterWeapon ("sg550", 23, "Sig SG-550 Sniper");
            RegisterWeapon ("m249", 24, "Fabrique Nationale M249 Para");
            RegisterWeapon ("shield", 25, "Tactical  Riot Shield");

            try
            {
               LoadData ();
            }
            catch (Exception e)
            {
               MessageBox.Show (e.Message + Localizer.T (" entry is not valid. Please fix config."), Localizer.T (CAPTION), MessageBoxButtons.OK, MessageBoxIcon.Error);
               Application.Exit ();
            }
            SetDefaults ();

            treeControl.Nodes[0].Toggle ();
         }

         public TreeNode FindNode (SkillLevel level, string name)
         {
            foreach (TreeNode n in treeControl.Nodes[0].Nodes)
            {
               if (n.Name == level.ToString ())
               {
                  foreach (TreeNode a in n.Nodes)
                  {
                     if (a.Name == name)
                        return a;
                  }
               }
            }
            return null;
         }

         public void GenerateNodeData (SkillLevel level, ref string output, ref int imageIndex)
         {
            switch (level)
            {
            case SkillLevel.Skill100:
               output += "Godlike";
               imageIndex = 8;
               break;

            case SkillLevel.Skill0:
               output += "Stupid";
               imageIndex = 1;
               break;

            case SkillLevel.Skill20:
               output += "Newbie";
               imageIndex = 2;
               break;

            case SkillLevel.Skill40:
               output += "Average";
               imageIndex = 3;
               break;

            case SkillLevel.Skill60:
               output += "Advanced";
               imageIndex = 4;
               break;

            case SkillLevel.Skill80:
               output += "Professional";
               imageIndex = 5;
               break;
            }
         }

         public void SetDefaults ()
         {
            loadData = defaultData;

            for (int i = 0; i < 6; i++)
            {
               for (int j = 0; j < NUM_SKILLENTRIES; j++)
               {
                  object val = 0;

                  switch (configValues[j])
                  {
                  case "MIN_DELAY":
                     val = loadData[i].minDelay;
                     break;

                  case "MAX_DELAY":
                     val = loadData[i].maxDelay;
                     break;

                  case "MIN_TURNSPEED":
                     val = loadData[i].minTurnspeed;
                     break;

                  case "MAX_TURNSPEED":
                     val = loadData[i].maxTurnspeed;
                     break;

                  case "HEADSHOT_ALLOW":
                     val = loadData[i].headshotAllow;
                     break;

                  case "HEAR_SHOOTTHRU":
                     val = loadData[i].hearShootThru;
                     break;

                  case "SEEN_SHOOTTHRU":
                     val = loadData[i].seenShootThru;
                     break;

                  case "RECOIL_AMOUNT":
                     val = loadData[i].recoilAmount;
                     break;
                  }
                  SetValue ((SkillLevel)i, configValues[j], val);
               }
            }
         }

         public void SetValue (SkillLevel level, string name, object value)
         {
            TreeNode node = FindNode (level, name);

            if (node != null)
            {
               node.Tag = value;

               switch (name)
               {
               case "MIN_TURNSPEED":
               case "MAX_TURNSPEED":
                  node.ToolTipText = Convert.ToString (node.Tag) + ".0";
                  break;

               case "MIN_DELAY":
               case "MAX_DELAY":
                  node.ToolTipText = (Convert.ToSingle (value) / 100).ToString ().Replace (',', '.');

                  if (node.ToolTipText.Length == 1 || node.ToolTipText.Length == 3)
                     node.ToolTipText += "0";

                  if (node.ToolTipText == "00")
                     node.ToolTipText = "0.0";
                  else if (node.ToolTipText == "10")
                     node.ToolTipText = "1.0";

                  break;

               default:
                  node.ToolTipText = Convert.ToString (node.Tag);
                  break;
               }
            }
         }

         public string GetValue (SkillLevel level, string name)
         {
            TreeNode node = FindNode (level, name);

            if (node != null)
               return Convert.ToString (node.Tag);

            return null;
         }

         public void AddSection (SkillLevel level)
         {
            string output = Localizer.T ("Definition for: ");
            string temp = "";
            int imageIndex = 0;

            GenerateNodeData (level, ref temp, ref imageIndex);
            output += " " + temp;

            TreeNode node = new TreeNode (output, imageIndex, imageIndex);
            node.Name = level.ToString ();

            for (int i = 0; i < NUM_SKILLENTRIES; i++)
            {
               if (i <= 3)
                  imageIndex = 6;
               else
                  imageIndex = 7;

               TreeNode n = new TreeNode (descriptionList[i], imageIndex, imageIndex);
               n.Name = configValues[i];

               node.Nodes.Add (n);
            }
            treeControl.Nodes[0].Nodes.Add (node);
         }
      };

      public GeneralManager generalManager;

      public class CvarItem
      {
         public string name;
         public string value;
         public string desc;


         public CvarItem (string name, string value, string desc)
         {
            this.name = name;
            this.desc = desc;
            this.value = value;
         }
      };

      public class HastableEx: Hashtable
      {
         public Hashtable descTab;

         public HastableEx ()
         {
            descTab = new Hashtable ();
         }

         public void AddVar (CvarItem item)
         {
            descTab.Add (item.name, item.desc);
            base.Add (item.name, item.value);
         }
      };

      public class ConfigManager
      {
         public HastableEx dataTab;
         private Main main;

         public string ValueForName (string name)
         {
            return dataTab[name].ToString ();
         }

         public void SetValue (string name, object value)
         {
            dataTab[name] = value;
         }

         public string DescForName (string name)
         {
            return dataTab.descTab[name].ToString ();
         }

         public ConfigManager (Main main)
         {
            dataTab = new HastableEx ();
            this.main = main;

            dataTab.AddVar (new CvarItem ("yb_autovacate", "0", Localizer.T ("Determines if one slot is keep free for player connection.")));
            dataTab.AddVar (new CvarItem ("yb_quota", "10", Localizer.T ("Determines number of bots playing on the server machine.")));
            dataTab.AddVar (new CvarItem ("yb_minskill", "95", Localizer.T ("Determines minimum bots skill, when creating bot with random parameters.")));
            dataTab.AddVar (new CvarItem ("yb_maxskill", "100", Localizer.T ("Determines maximum bots skill, when creating bot with random parameters.")));
            dataTab.AddVar (new CvarItem ("yb_followuser", "3", Localizer.T ("Determines number of bots that can follow player, that used Follow me radio message.")));
            dataTab.AddVar (new CvarItem ("yb_timersound", "0.5", Localizer.T ("Determines time interval, used with bot hearing system.")));
            dataTab.AddVar (new CvarItem ("yb_timerpickup", "0.2", Localizer.T ("Determines time interval, used to check for items around the bot.")));
            dataTab.AddVar (new CvarItem ("yb_timergrenade", "0.5", Localizer.T ("Determines time interval, used to check for grenade throw.")));
            dataTab.AddVar (new CvarItem ("yb_chat", "1", Localizer.T ("Determines if bots able to chat each-other while they are dead.")));
            dataTab.AddVar (new CvarItem ("yb_synth", "1", Localizer.T ("Determines if bot dll able to comment some of the commands, by halflife speech engine.")));
            dataTab.AddVar (new CvarItem ("yb_knifemode", "0", Localizer.T ("Determines mode, where bots able to fight only on knives.")));
            dataTab.AddVar (new CvarItem ("yb_skilltags", "0", Localizer.T ("Determines, if skill value for bot will be appended to it name, after bot creation.")));
            dataTab.AddVar (new CvarItem ("yb_stop", "0", Localizer.T ("Determines, if all bots on the server stop function, while this variable set to 1.")));
            dataTab.AddVar (new CvarItem ("yb_thruwalls", "1", Localizer.T ("Determines, if bots able to shoot through wall when suspectiong an enemy.")));
            dataTab.AddVar (new CvarItem ("yb_votes", "1", Localizer.T ("Determines, if bots can vote for kick, against it's teamkiller, or for map change.")));
            dataTab.AddVar (new CvarItem ("yb_spray", "1", Localizer.T ("Determines, if bots are able to spray graphical logos in a game.")));
            dataTab.AddVar (new CvarItem ("yb_buy", "1", Localizer.T ("Determines, if bots able to purchase weapon and items at start of the each round.")));
            dataTab.AddVar (new CvarItem ("yb_debug", "0", Localizer.T ("Determines level for bot debugging engine. For developers and betatesting team only.")));
            dataTab.AddVar (new CvarItem ("yb_turnmethod", "3", Localizer.T ("Determines method for aiming of bots.")));
            dataTab.AddVar (new CvarItem ("yb_prefix", "", Localizer.T ("Determines prefix which will be added to start of the bot name, at the creation time.")));            
            dataTab.AddVar (new CvarItem ("yb_language", "en", Localizer.T ("Determines the language in which all bot menus and comments are displayed.")));
            dataTab.AddVar (new CvarItem ("yb_password", "0", Localizer.T ("Determines password for accessing bot menu and commands on the server running yapb.")));
            dataTab.AddVar (new CvarItem ("yb_passwordkey", "0", Localizer.T ("Determines pasword key for setinfo information to acccess to the bot menu and command.")));
            dataTab.AddVar (new CvarItem ("yb_voicepath", "sound/radio/bot", Localizer.T ("Determines path to the Official CSBot's chatter files.")));
            dataTab.AddVar (new CvarItem ("yb_tkpunish", "1", Localizer.T ("Determines if bot's teamkiller will be punshied for his heroic action.")));
            dataTab.AddVar (new CvarItem ("yb_communication", "2", Localizer.T ("Determines method for communication engine for bots.")));
            dataTab.AddVar (new CvarItem ("yb_economics", "1", Localizer.T ("Determines if bots are able to use team-economics to save money.")));
            dataTab.AddVar (new CvarItem ("yb_forceteam", "any", Localizer.T ("Determines the team, in which all creating bots will be forced to join.")));
            dataTab.AddVar (new CvarItem ("yb_restrweapons", "ump45;p90;elite;tmp;mac10;m3;xm1014", Localizer.T ("Determines list of the weapons, that are banned for buying and picking up.")));
            dataTab.AddVar (new CvarItem ("yb_dangerfactor", "800", Localizer.T ("Determines danger factor for bots experience system, higher values resulting more tactical bots.")));
            dataTab.AddVar (new CvarItem ("yb_dontshoot", "0", Localizer.T ("Determines if bots are able to shoot their enemies.")));
            dataTab.AddVar (new CvarItem ("yb_hardcoremode", "1", Localizer.T ("Determines if bots can only be running in hard mode. Enable only if you playing very well.")));
            dataTab.AddVar (new CvarItem ("yb_autofollow_percent", "50", Localizer.T ("Determines percent of bots that can autofollow human player at start of each round.")));
            dataTab.AddVar (new CvarItem ("yb_logotypes", "{biohaz;{graf004;{graf005;{lambda06;{target;{hand1", Localizer.T ("Determines the list of logos used for bot spray logo action.")));
            dataTab.AddVar (new CvarItem ("yb_msecmethod", "1", Localizer.T ("Determines method for calculating the msec value.")));
				dataTab.AddVar (new CvarItem ("yb_loglevel", "2", Localizer.T ("Determines log level for bots messages. Critical errors will be logged anyway.")));
			
			}

         public HastableEx Entries
         {
            get
            {
               return dataTab;
            }
         }

         public void LoadData ()
         {
            string cfg = configPath + "yapb.cfg";

            using (TextReader reader = new StreamReader (File.OpenRead (cfg)))
            {
               for (; ; )
               {
                  string line = reader.ReadLine ();

                  if (line == null)
                     break;

                  if ((line.Contains ("//") && line[0] == '/' && (line[1] == '/')) || line == "\n" || line == "\r" || line == "\0")
                     continue;

                  line.Trim ('\n', '\r');

                  string[] list = line.Split (' ');

                  if (list.Length < 2)
                     continue;

                  list[1] = list[1].Trim ('\r', '\n', '\"');

                  try
                  {
                     main.SetVar (list[0], list[1]);
                  }
                  catch (Exception e)
                  {
                     MessageBox.Show (main, Localizer.T ("Something weird, with your config. Probably some values are messed. Config loaded not completely.\n\nException:\n") + e.Message + e.StackTrace + "", Localizer.T (CAPTION), MessageBoxButtons.OK, MessageBoxIcon.Error);
                  }
               }
            }
         }


         internal string GenerateDefault (string cvar, string value)
         {
            return "// " + dataTab.descTab[cvar] + "\n" + cvar + " \"" + value + "\"\n\n";
         }

         internal string BooleanValue (CheckBox ctl)
         {
            return ctl.Checked ? "1" : "0";
         }

         internal string BarValue (HScrollBar ctl)
         {
            return ctl.Value.ToString ();
         }

         internal string FloatValue (HScrollBar ctl)
         {
            string value = ctl.Value.ToString ();
            value = value.Insert (0, "0.");

            if (value == "0.10")
               value = "1.0";

            return value;
         }

         /// <summary>
         ///  Writes output, rewriting it completely.
         /// </summary>
         public void WriteOutputConfig (string toFile, string restrWeapons)
         {
            string configHead = CONFIG_HDR + "\n";

            configHead += GenerateDefault ("yb_autofollow_percent", BarValue (main.Autofollow));
            configHead += GenerateDefault ("yb_autovacate", BooleanValue (main.AutoVacate));
            configHead += GenerateDefault ("yb_quota", BarValue (main.Quota));
            configHead += GenerateDefault ("yb_minskill", BarValue (main.MinSkill));
            configHead += GenerateDefault ("yb_maxskill", BarValue (main.MaxSkill));
            configHead += GenerateDefault ("yb_dangerfactor", BarValue (main.Factor));
            configHead += GenerateDefault ("yb_followuser", BarValue (main.FollowUser));
            configHead += GenerateDefault ("yb_timersound", FloatValue (main.SoundTimer));
            configHead += GenerateDefault ("yb_timerpickup", FloatValue (main.PickupTimer));
            configHead += GenerateDefault ("yb_timergrenade", FloatValue (main.GrenadeTimer));
            configHead += GenerateDefault ("yb_chat", BooleanValue (main.Chat));
            configHead += GenerateDefault ("yb_synth", BooleanValue (main.Synth));
            configHead += GenerateDefault ("yb_knifemode", BooleanValue (main.KnifeMode));
            configHead += GenerateDefault ("yb_skilltags", BooleanValue (main.SkillTags));
            configHead += GenerateDefault ("yb_stop", BooleanValue (main.Stop));
            configHead += GenerateDefault ("yb_thruwalls", BooleanValue (main.ThruWall));
            configHead += GenerateDefault ("yb_votes", BooleanValue (main.Voting));
            configHead += GenerateDefault ("yb_spray", BooleanValue (main.Spraying));
            configHead += GenerateDefault ("yb_buy", BooleanValue (main.Buying));
            configHead += GenerateDefault ("yb_debug", "0");
            configHead += GenerateDefault ("yb_tkpunish", BooleanValue (main.Punishment));
            configHead += GenerateDefault ("yb_dontshoot", BooleanValue (main.Dontshoot));
            configHead += GenerateDefault ("yb_hardcoremode", BooleanValue (main.HardcoreMode));
            configHead += GenerateDefault ("yb_economics", BooleanValue (main.Economics));
            configHead += GenerateDefault ("yb_prefix", main.Prefix.Text);
            configHead += GenerateDefault ("yb_passwordkey", main.PasswordKey.Text);
            configHead += GenerateDefault ("yb_password", main.Password.Text);
            configHead += GenerateDefault ("yb_voicepath", main.ChatterPath.Text);

            if (restrWeapons == "")
               configHead += GenerateDefault ("yb_restrweapons", "xm1014;m3");
            else
               configHead += GenerateDefault ("yb_restrweapons", restrWeapons);


            if (main.Logotypes.Text == "")
               configHead += GenerateDefault ("yb_logotypes", "{biohaz;{graf004;{graf005;{lambda06;{target;{hand1");
            else
               configHead += GenerateDefault ("yb_logotypes", main.Logotypes.Text);

            if (main.Turnmethod.SelectedIndex != -1)
               configHead += GenerateDefault ("yb_turnmethod", (main.Turnmethod.SelectedIndex + 1).ToString ());
            else
               configHead += GenerateDefault ("yb_turnmethod", "3");

            if (main.Msec.SelectedIndex != -1)
               configHead += GenerateDefault ("yb_msecmethod", (main.Msec.SelectedIndex + 1).ToString ());
            else
               configHead += GenerateDefault ("yb_msecmethod", "1");

				if (main.Loglevel.SelectedIndex != -1)
					configHead += GenerateDefault ("yb_loglevel", main.Loglevel.SelectedIndex.ToString ());
				else
					configHead += GenerateDefault ("yb_loglevel", "1");

            if (main.Language.SelectedIndex != -1)
            {
               string langCode = main.Language.Text;
               langCode = langCode.Remove (0, 1);
               langCode = langCode.Remove (langCode.IndexOf (']'), langCode.Length - langCode.IndexOf (']'));

               configHead += GenerateDefault ("yb_language", langCode);
            }
            else
               configHead += GenerateDefault ("yb_language", "en");

            if (main.Communcation.SelectedIndex != -1)
               configHead += GenerateDefault ("yb_communication", main.Communcation.SelectedIndex.ToString ());
            else
               configHead += GenerateDefault ("yb_communication", "2");

            if (main.ForcedTeam.SelectedIndex != -1)
            {
               if (main.ForcedTeam.SelectedIndex == 0)
                  configHead += GenerateDefault ("yb_forceteam", "t");
               else if (main.ForcedTeam.SelectedIndex == 1)
                  configHead += GenerateDefault ("yb_forceteam", "ct");
               else
                  configHead += GenerateDefault ("yb_forceteam", "any");
            }
            else
               configHead += GenerateDefault ("yb_forceteam", "any");

            WriteFile_ANSI (toFile, configHead);
         }
      };

      public ConfigManager configManager;

      public void WeaponListInvalidate (Personality type)
      {
         WeaponList.Items.Clear ();
         generalManager.Invalidate (type);
      }

      public void SetVar (string cvar, string value)
      {
         cvar = cvar.Remove (0, 3);

			if (cvar.Contains ("timer"))
			{
				int val = 0;

				if (value.Contains ("0."))
					val = Convert.ToInt32 (value.Remove (0, 2));

				if (cvar == "timersound")
					SoundTimer.Value = val;
				else if (cvar == "timergrenade")
					GrenadeTimer.Value = val;
				else if (cvar == "timerpickup")
					PickupTimer.Value = val;
			}
			else if (cvar == "autofollow_percent")
				Autofollow.Value = Convert.ToInt32 (value);
			else if (cvar == "quota")
				Quota.Value = Convert.ToInt32 (value);
			else if (cvar == "minskill")
				MinSkill.Value = Convert.ToInt32 (value);
			else if (cvar == "maxskill")
				MaxSkill.Value = Convert.ToInt32 (value);
			else if (cvar == "followuser")
				FollowUser.Value = Convert.ToInt32 (value);
			else if (cvar == "hardcoremode")
				HardcoreMode.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "dontshoot")
				Dontshoot.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "buy")
				Buying.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "spray")
				Spraying.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "votes")
				Voting.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "thruwalls")
				ThruWall.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "skilltags")
				SkillTags.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "autovacate")
				AutoVacate.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "knifemode")
				KnifeMode.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "chat")
				Chat.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "synth")
				Synth.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "stop")
				Stop.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "economics")
				Economics.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "communication" && value != "")
				Communcation.SelectedIndex = Convert.ToInt32 (value);
			else if (cvar == "loglevel" && value != "")
				Loglevel.SelectedIndex = Convert.ToInt32 (value);
			else if (cvar == "turnmethod" && value != "")
				Turnmethod.SelectedIndex = Convert.ToInt32 (value) - 1;
			else if (cvar == "msecmethod" && value != "")
				Msec.SelectedIndex = Convert.ToInt32 (value) - 1;
			else if (cvar == "dangerfactor")
				Factor.Value = Convert.ToInt32 (value);
			else if (cvar == "tkpunish")
				Punishment.Checked = Convert.ToInt32 (value) > 0;
			else if (cvar == "passwordkey")
				PasswordKey.Text = value;
			else if (cvar == "password")
				Password.Text = value;
			else if (cvar == "logotypes")
				Logotypes.Text = value;
			else if (cvar == "voicepath")
				ChatterPath.Text = value;
			else if (cvar == "prefix")
				Prefix.Text = value;
			else if (cvar == "dagerfactor")
				Factor.Value = Convert.ToInt32 (value);
			else if (cvar == "forceteam")
			{
				if (value == "any")
					ForcedTeam.SelectedIndex = 2;
				else if (value == "ct")
					ForcedTeam.SelectedIndex = 1;
				else if (value == "t")
					ForcedTeam.SelectedIndex = 0;
			}
			else if (cvar == "language")
			{
				for (int i = 0; i < Language.Items.Count; i++)
				{
					if (Language.Items[i].ToString ().Contains (value))
					{
						Language.SelectedIndex = i;
						break;
					}
				}
			}
			else if (cvar == "restrweapons")
			{
				ArrayList list = new ArrayList (NUM_WEAPONS);
				string[] items = value.Split (';');

				for (int i = 0; i < items.Length; i++)
					list.Add (items[i]);

				foreach (WeaponItem item in generalManager.Entries)
				{
					if (list.Contains (item.Name))
					{
						for (int i = 0; i < RestrictedWeapons.Items.Count; i++)
						{
							if (RestrictedWeapons.Items[i].ToString () == "[" + item.Name + "] " + item.Desc)
								RestrictedWeapons.SetItemChecked (i, true);
						}
					}
				}
			}
      }

      public string CurrentRestrictedItems ()
      {
         string listItems = "";

         foreach (int it in RestrictedWeapons.CheckedIndices)
         {
            string res = "";

            foreach (WeaponItem item in generalManager.Entries)
            {
               if ("[" + item.Name + "] " + item.Desc == RestrictedWeapons.Items[it].ToString ())
               {
                  res = item.Name;
                  break;
               }
            }
            listItems += res + ";";
         }

         if (listItems.Length > 0)
            listItems = listItems.Trim (';', ' ');

         return listItems;
      }

      public void InitMainConfigControls ()
      {
         FillRestTab ();

         DirectoryInfo dirInfo = new DirectoryInfo (configPath + "language");
         FileInfo[] list = dirInfo.GetFiles ();
         ArrayList usedLangCodes = new ArrayList (32);

         foreach (FileInfo it in list)
         {
            if (it.Extension != ".cfg")
               continue;

            string code = it.Name.Substring (0, 3);

            if (code.Contains ("_"))
               code = code.Remove (code.IndexOf ('_'));

            if (usedLangCodes.Contains (code))
               continue;

            usedLangCodes.Add (code);

            switch (code)
            {
            case "ru":
               code = Localizer.T ("[ru] Russian Federation  (Russia)");
               break;

            case "en":
               code = Localizer.T ("[en] United States (English)");
               break;

            case "chs":
               code = Localizer.T ("[chs] China (Simplified Chinese)");
               break;
               
            case "pl":
               code = Localizer.T ("[pl] Poland (Polish)");
               break;
            };
            Language.Items.Add (code);
         }
      }

      static public string botPath;
      static public string configPath;

      public Main ()
      {
         InitializeComponent ();
         Text += " v" + VERSION;
      }

      private void AckLoad (object sender, EventArgs e)
      {
         generalManager = new GeneralManager (SkillView, WeaponList, PerSelect);
         configManager = new ConfigManager (this);

         WeaponDown.Enabled = false;
         WeaponUp.Enabled = false;
         ScrollChange.Enabled = false;

         PerSelect.SelectedIndex = 1;
         InitMainConfigControls ();
         configManager.LoadData ();

         if (Application.VisualStyleState == VisualStyleState.NoneEnabled)
         {
            ScrollChange.BackColor = System.Drawing.SystemColors.Control;
            GroupBox.BackColor = System.Drawing.SystemColors.Control;
         }
      }

      private void ExitClick (object sender, EventArgs e)
      {
         Application.Exit ();
      }

      private void AboutBox (object sender, EventArgs e)
      {
         About dlg = new About ();
         dlg.ShowDialog (this);
      }

      private void PerSelect_SelectedIndexChanged (object sender, EventArgs e)
      {
         WeaponListInvalidate (GetCurrent ());
      }

      internal protected Personality GetCurrent ()
      {
         Personality type = Personality.Normal;

         if (PerSelect.SelectedIndex == 0)
            type = Personality.Rusher;
         else if (PerSelect.SelectedIndex == 1)
            type = Personality.Normal;
         else if (PerSelect.SelectedIndex == 2)
            type = Personality.Careful;

         return type;
      }

      private void WeaponListMenu_Opening (object sender, CancelEventArgs e)
      {
         if (WeaponList.SelectedItems.Count == 0)
            e.Cancel = true;
      }

      public void ScrollValue (HScrollBar ctl)
      {
         if (ctl.Tag == null)
            return;

         if (ctl.Tag.ToString ().Contains ("timer"))
         {
            GlobalPopup.SetToolTip (ctl, configManager.DescForName (ctl.Tag.ToString ()) + Localizer.T ("\r\nValue for ") + ctl.Tag.ToString () + ": " + (ctl.Value.ToString () == "10" ? "1.0" : "0." + ctl.Value.ToString ()) + ".");
            return;
         }

         GlobalPopup.SetToolTip (ctl, configManager.DescForName (ctl.Tag.ToString ()) + Localizer.T ("\r\nValue for ") + ctl.Tag.ToString () + ": " + ctl.Value.ToString () + ".");
      }

      public void CheckboxValue (CheckBox ctl)
      {
         if (ctl.Tag == null)
            return;

         GlobalPopup.SetToolTip (ctl, configManager.DescForName (ctl.Tag.ToString ()) + Localizer.T ("\r\nValue for ") + ctl.Tag.ToString () + ": " + (ctl.Checked ? Localizer.T ("Enabled") : Localizer.T ("Disabled")) + ".");
      }

      public void ComboBoxValue (ComboBox ctl)
      {
         if (ctl.Tag == null)
            return;

         GlobalPopup.SetToolTip (ctl, configManager.DescForName (ctl.Tag.ToString ()) + Localizer.T ("\r\nValue for ") + ctl.Tag.ToString () + ": " + (ctl.Text == "" ? Localizer.T ("None") : ctl.Text));
      }

      public void TextBoxValue (TextBox ctl)
      {
         if (ctl.Tag == null)
            return;

         GlobalPopup.SetToolTip (ctl, configManager.DescForName (ctl.Tag.ToString ()) + Localizer.T ("\r\nValue for ") + ctl.Tag.ToString () + ": " + (ctl.Text == "" ? Localizer.T ("None") : ctl.Text));
      }

      public void RestrictionValue (CheckedListBox ctl)
      {
         if (ctl.Tag == null)
            return;

         GlobalPopup.SetToolTip (ctl, configManager.DescForName (ctl.Tag.ToString ()) + Localizer.T ("\r\nValue for ") + ctl.Tag.ToString () + ": " + CurrentRestrictedItems ());
      }

      public void FillRestTab ()
      {
         RestrictedWeapons.Items.Clear ();

         foreach (WeaponItem it in generalManager.Entries)
            RestrictedWeapons.Items.Add ("[" + it.Name + "] " + it.Desc);
      }

      internal protected bool ValidateKnife (ListViewItem item)
      {
         WeaponItem it = generalManager.ItemToEntry (GetCurrent (), item, -1);

         if (it.Name == "knife")
         {
            MessageBox.Show (this, Localizer.T ("Knife priority cannot be changed."), Localizer.T (CAPTION), MessageBoxButtons.OK, MessageBoxIcon.Error);
            return false;
         }
         return true;
      }

      internal protected void RegularClick (bool up)
      {
         if (WeaponList.SelectedItems.Count == 0)
            return;

         ListViewItem item = WeaponList.SelectedItems[0];

         if (!ValidateKnife (item))
            return;

         int index = item.Index;
         int to = (up ? (index + 1) : (index - 1));

         if (to == 0)
         {
            MessageBox.Show (this, Localizer.T ("Cannot move item to knife position. Knife priority cannot be changed."), Localizer.T (CAPTION), MessageBoxButtons.OK, MessageBoxIcon.Error);
            return;
         }
         if (to < 0 || to >= NUM_WEAPONS)
         {
            MessageBox.Show (this, Localizer.T ("Cannot move item to this position. Index is out of range."), Localizer.T (CAPTION), MessageBoxButtons.OK, MessageBoxIcon.Error);
            return;
         }

         WeaponList.Items.Remove (item);
         WeaponList.Items.Insert (to, item);

         generalManager.RebuildTable (GetCurrent ());
         WeaponList.SelectedItems[0].Selected = true;
         WeaponList.SelectedItems[0].EnsureVisible ();
      }

      private void WeaponButton_UpClick (object sender, EventArgs e)
      {
         RegularClick (false);
      }

      private void WeaponButton_DownClick (object sender, EventArgs e)
      {
         RegularClick (true);
      }

      private void WeaponPercent_Click (object sender, EventArgs e)
      {
         PercentEdit dlg = new PercentEdit (GeneralManager.Percentage);
         dlg.ShowDialog (this);
      }

      private void WeaponEconomics_Click (object sender, EventArgs e)
      {
         EconomicsEdit dlg = new EconomicsEdit (GeneralManager.Economics);
         dlg.ShowDialog (this);
      }

      private void ListChange (object sender, EventArgs e)
      {
         WeaponDown.Enabled = true;
         WeaponUp.Enabled = true;
      }

      internal TreeNode currentNode;

      private void ValueChanged (object sender, EventArgs e)
      {
         ChangeDisplay.Text = "";

         switch (currentNode.Name)
         {
         case "HEADSHOT_ALLOW":
         case "HEAR_SHOOTTHRU":
         case "SEEN_SHOOTTHRU":
         case "RECOIL_AMOUNT":
            ChangeDisplay.Text = ScrollChange.Value.ToString ();

            for (int i = 0; i < 6; i++)
            {
               if (currentNode.Parent.Name == ((SkillLevel)i).ToString ())
               {
                  generalManager.SetValue ((SkillLevel)i, currentNode.Name, ScrollChange.Value.ToString ());
                  //break;
               }
            }
            break;

         case "MIN_TURNSPEED":
         case "MAX_TURNSPEED":
            ChangeDisplay.Text = ScrollChange.Value.ToString () + ".0";

            for (int i = 0; i < 6; i++)
            {
               if (currentNode.Parent.Name == ((SkillLevel)i).ToString ())
               {
                  generalManager.SetValue ((SkillLevel)i, currentNode.Name, ScrollChange.Value.ToString ());
                  //break;
               }
            }
            break;

         case "MIN_DELAY":
         case "MAX_DELAY":
            ChangeDisplay.Text = (Convert.ToSingle (ScrollChange.Value) / 100).ToString ().Replace (',', '.');

            if (ChangeDisplay.Text.Length == 1 || ChangeDisplay.Text.Length == 3)
               ChangeDisplay.Text += "0";

            if (ChangeDisplay.Text == "00")
               ChangeDisplay.Text = "0.0";
            else if (ChangeDisplay.Text == "10")
               ChangeDisplay.Text = "1.0";

            for (int i = 0; i < 6; i++)
            {
               if (currentNode.Parent.Name == ((SkillLevel)i).ToString ())
               {
                  generalManager.SetValue ((SkillLevel)i, currentNode.Name, (ScrollChange.Value.ToString ()));
                  // break;
               }
            }
            break;
         }
      }

      private void SkillView_NodeMouseClick (object sender, TreeNodeMouseClickEventArgs e)
      {
         if (!e.Node.IsExpanded && e.Node.GetNodeCount (true) == 0)
         {
            ScrollChange.Enabled = true;

            string text = e.Node.Parent.Text;
            text = text.Remove (0, 16);

            ChangeDesc.Text = e.Node.Text + ": " + text;
            ChangeDisplay.Text = e.Node.Tag.ToString ();

            currentNode = e.Node;

            switch (e.Node.Name)
            {
            case "HEADSHOT_ALLOW":
            case "HEAR_SHOOTTHRU":
            case "SEEN_SHOOTTHRU":
               ScrollChange.SetRange (0, 100);
               goto case "DEF";

            case "MIN_TURNSPEED":
            case "MAX_TURNSPEED":
               ScrollChange.SetRange (20, 120);
               goto case "FLT2";

            case "RECOIL_AMOUNT":
               ScrollChange.SetRange (20, 50);
               goto case "DEF";

            case "MIN_DELAY":
            case "MAX_DELAY":
               ScrollChange.SetRange (0, 100);
               goto case "FLT";

            case "FLT":
               ScrollChange.SmallChange = ScrollChange.LargeChange = 1;
               ScrollChange.Value = Convert.ToInt32 (e.Node.Tag.ToString ());

               if (!ChangeDisplay.Text.Contains ("."))
               {
                  string val = (Convert.ToSingle (ChangeDisplay.Text) / 100).ToString ().Replace (',', '.');

                  if (val.Length == 1 || val.Length == 3)
                     val += "0";

                  if (val == "00")
                     val = "0.0";
                  else if (val == "10")
                     val = "1.0";

                  ChangeDisplay.Text = val;
               }

               break;

            case "FLT2":
               ScrollChange.SmallChange = ScrollChange.LargeChange = 1;
               ScrollChange.Value = Convert.ToInt32 (e.Node.Tag.ToString ());

               if (!ChangeDisplay.Text.Contains ("."))
                  ChangeDisplay.Text += ".0";
               break;


            case "DEF":
               ScrollChange.SmallChange = ScrollChange.LargeChange = 1;
               ScrollChange.Value = Convert.ToInt32 (ChangeDisplay.Text);

               break;
            }
         }
      }

      private void MainMenu_KeyDown (object sender, KeyEventArgs e)
      {
         if (e.Shift && e.KeyCode == Keys.Down)
            RegularClick (true);
         else if (e.Shift && e.KeyCode == Keys.Up)
            RegularClick (false);
      }

      private void DefaultReset_Click (object sender, EventArgs e)
      {
         if (MessageBox.Show (this, Localizer.T ("Are you sure you want reset skill settings ?"), Localizer.T (CAPTION), MessageBoxButtons.YesNo, MessageBoxIcon.Question) == DialogResult.Yes)
         {
            generalManager.loadData = GeneralManager.defaultData;
            generalManager.SetDefaults ();

            ChangeDisplay.Text = "";
            ScrollChange.Enabled = false;
         }
      }

      private void ToggleTree_Click (object sender, EventArgs e)
      {
         int numExpanded = 0;

         for (int i = 0; i < 6; i++)
         {
            if (SkillView.Nodes[0].Nodes[i].IsExpanded)
               numExpanded++;
         }

         if (numExpanded > 3)
         {
            for (int i = 0; i < 6; i++)
               SkillView.Nodes[0].Nodes[i].Collapse ();
         }
         else
         {
            for (int i = 0; i < 6; i++)
               SkillView.Nodes[0].Nodes[i].Expand ();
         }
      }

      private void SkillView_BeforeSelect (object sender, TreeViewCancelEventArgs e)
      {
         if (e.Action == TreeViewAction.ByKeyboard)
            e.Cancel = true;
      }

      private void WebsiteLink_LinkClicked (object sender, LinkLabelLinkClickedEventArgs e)
      {
         e.Link.Visited = true;
         UnmanagedUtils.ShellExecute (IntPtr.Zero, "open", "http://yapb.bots-united.com", null, null, 1);
      }

      private void SaveExit_Click (object sender, EventArgs e)
      {
         generalManager.WriteOutputConfig (configPath + "general.cfg");
         configManager.WriteOutputConfig (configPath + "yapb.cfg", CurrentRestrictedItems ());

         Hide ();

         FakeLoad dlg = new FakeLoad ();
         dlg.SleepAmount = 1;
         dlg.Desc = Localizer.T ("Writing configs...");

         dlg.ShowDialog (this);

         ExitClick (sender, e);
      }

      private void ReloadMain (object sender, EventArgs e)
      {
         configManager.LoadData ();
      }

      private void ReloadGeneral (object sender, EventArgs e)
      {
         generalManager.LoadData ();
         generalManager.SetDefaults ();

         for (int i = 0; i < NUM_PERSONALITIES; i++)
            WeaponListInvalidate ((Personality)i);

         ChangeDisplay.Text = "";
         ScrollChange.Enabled = false;
      }

      private void WriteGeneral (object sender, EventArgs e)
      {
         generalManager.WriteOutputConfig (configPath + "general.cfg");

      }

      private void WriteMain (object sender, EventArgs e)
      {
         configManager.WriteOutputConfig (configPath + "yapb.cfg", CurrentRestrictedItems ());
      }

      private void ScrollChanger (object sender, EventArgs e)
      {
         ScrollValue ((HScrollBar)sender);
      }

      private void ResetRestriction_Click (object sender, EventArgs e)
      {
         RestrictedWeapons.Items.Clear ();

         FillRestTab ();
      }

      private void CheckBoxChanger (object sender, EventArgs e)
      {
         CheckboxValue ((CheckBox)sender);
      }

      private void ComboBoxChanger (object sender, EventArgs e)
      {
         ComboBoxValue ((ComboBox)sender);
      }

      private void TextBoxChanged (object sender, EventArgs e)
      {
         TextBoxValue ((TextBox)sender);
      }

      private void RestrictionChanger (object sender, EventArgs e)
      {
         RestrictionValue ((CheckedListBox)sender);
      }

      private void RestrictionChanger (object sender, ItemCheckEventArgs e)
      {
         RestrictionValue ((CheckedListBox)sender);
      }

      private void ExitWithSave (object sender, EventArgs e)
      {
         SaveExit_Click (null, null);
      }

      private void BaseImage_Click (object sender, EventArgs e)
      {
         UnmanagedUtils.ShellExecute (IntPtr.Zero, "open", "http://cs-6hqp.ukrnic.com/", null, null, 1);
      }

      private void WeaponList_SetHighestPriority (object sender, EventArgs e)
      {
         if (WeaponList.SelectedItems.Count == 0)
            return;

         ListViewItem item = WeaponList.SelectedItems[0];

         if (!ValidateKnife (item))
            return;

         WeaponList.SelectedItems[0].Remove ();

         ListViewItem item2 = WeaponList.Items.Insert (NUM_WEAPONS - 1, item);

         generalManager.RebuildTable (GetCurrent ());

         item2.Selected = true;
         item2.EnsureVisible ();
      }

      private void WeaponList_SetLowestPriority (object sender, EventArgs e)
      {
         if (WeaponList.SelectedItems.Count == 0)
            return;

         ListViewItem item = WeaponList.SelectedItems[0];

         if (!ValidateKnife (item))
            return;

         WeaponList.SelectedItems[0].Remove ();

         ListViewItem item2 = WeaponList.Items.Insert (1, item);

         generalManager.RebuildTable (GetCurrent ());

         item2.Selected = true;
         item2.EnsureVisible ();
      }
   };

   public static class UnmanagedUtils
   {
      // P-Invoke Stuff

      [DllImport ("shell32.dll", CharSet = CharSet.Auto)]
      public extern static int ShellExecute (IntPtr hwnd, [MarshalAs (UnmanagedType.LPTStr)] string op, [MarshalAs (UnmanagedType.LPTStr)] string file, [MarshalAs (UnmanagedType.LPTStr)] string parms, [MarshalAs (UnmanagedType.LPTStr)] string dir, int showCmd);

      [DllImport ("kernel32.dll", CharSet = CharSet.Auto)]
      public extern static int Sleep (uint ms);

      [DllImport ("msvcrt.dll")]
      public extern static double atof (string val);
   };
};